@@ -40,9 +40,9 @@ import (
 // These come from just running the code.
 // TODO: Auto-generate these in a much more sane way.
 const (
-	expectedLayerDigest    = "sha256:9a98de6b2015d531559791e60518fd376ddc62d3062ee4f691b223c06175dbef"
-	expectedConfigDigest   = "sha256:1d043a5807e0ca5bcde233f14a79928f9aa5ccecdd4a8e4bf4cdd0b557090f91"
-	expectedManifestDigest = "sha256:8b8f01aa9da2942d5dde5032b501004957f060b027c27162346972cff9220200"
+	expectedLayerDigest    = "sha256:96338a7c847bc582c82e4962a4285afcaf568e3913b0542b8745be27a418a806"
+	expectedConfigDigest   = "sha256:ddcc2a93d5b0bcdcb571431c3607d84abe3752406f7c631a898340e6e7e61ed0"
+	expectedManifestDigest = "sha256:0e8b342d2b01241b3f0197d0210ed5c0012d01817881defc1464e000f5b08f4d"
 )
 
 func setup(t *testing.T, dir string) (cas.Engine, ispec.Descriptor) {
@@ -62,9 +62,9 @@ func setup(t *testing.T, dir string) (cas.Engine, ispec.Descriptor) {
 	tw := tar.NewWriter(&buffer)
 	data := []byte("some contents")
 	tw.WriteHeader(&tar.Header{
+		Typeflag: tar.TypeReg,
 		Name:     "test",
 		Mode:     0644,
-		Typeflag: tar.TypeRegA,
 		Size:     int64(len(data)),
 	})
 	tw.Write(data)
@@ -215,7 +215,7 @@ func TestMutateAdd(t *testing.T) {
 	buffer := bytes.NewBufferString("contents")
 
 	// Add a new layer.
-	if err := mutator.Add(context.Background(), buffer, ispec.History{
+	if err := mutator.Add(context.Background(), buffer, &ispec.History{
 		Comment: "new layer",
 	}); err != nil {
 		t.Fatalf("unexpected error adding layer: %+v", err)
@@ -295,7 +295,7 @@ func TestMutateAddNonDistributable(t *testing.T) {
 	buffer := bytes.NewBufferString("contents")
 
 	// Add a new layer.
-	if err := mutator.AddNonDistributable(context.Background(), buffer, ispec.History{
+	if err := mutator.AddNonDistributable(context.Background(), buffer, &ispec.History{
 		Comment: "new layer",
 	}); err != nil {
 		t.Fatalf("unexpected error adding layer: %+v", err)
@@ -374,7 +374,7 @@ func TestMutateSet(t *testing.T) {
 	// Change the config
 	if err := mutator.Set(context.Background(), ispec.ImageConfig{
 		User: "changed:user",
-	}, Meta{}, nil, ispec.History{
+	}, Meta{}, nil, &ispec.History{
 		Comment: "another layer",
 	}); err != nil {
 		t.Fatalf("unexpected error adding layer: %+v", err)
@@ -429,6 +429,71 @@ func TestMutateSet(t *testing.T) {
 	}
 }
 
+func TestMutateSetNoHistory(t *testing.T) {
+	dir, err := ioutil.TempDir("", "umoci-TestMutateSetNoHistory")
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer os.RemoveAll(dir)
+
+	engine, fromDescriptor := setup(t, dir)
+	defer engine.Close()
+
+	mutator, err := New(engine, casext.DescriptorPath{Walk: []ispec.Descriptor{fromDescriptor}})
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// Change the config
+	if err := mutator.Set(context.Background(), ispec.ImageConfig{
+		User: "changed:user",
+	}, Meta{}, nil, nil); err != nil {
+		t.Fatalf("unexpected error adding layer: %+v", err)
+	}
+
+	newDescriptor, err := mutator.Commit(context.Background())
+	if err != nil {
+		t.Fatalf("unexpected error committing changes: %+v", err)
+	}
+
+	if newDescriptor.Descriptor().Digest == fromDescriptor.Digest {
+		t.Fatalf("new and old descriptors are the same!")
+	}
+
+	mutator, err = New(engine, newDescriptor)
+	if err != nil {
+		t.Fatal(err)
+	}
+
+	// Cache the data to check it.
+	if err := mutator.cache(context.Background()); err != nil {
+		t.Fatalf("unexpected error getting cache: %+v", err)
+	}
+
+	// Check digests are different.
+	if mutator.manifest.Config.Digest == expectedConfigDigest {
+		t.Errorf("manifest.Config.Digest is the same!")
+	}
+
+	// Check layer was not added.
+	if len(mutator.manifest.Layers) != 1 {
+		t.Errorf("manifest.Layers was updated")
+	}
+
+	// Check config was also modified.
+	if len(mutator.config.RootFS.DiffIDs) != 1 {
+		t.Errorf("config.RootFS.DiffIDs was updated")
+	}
+	if mutator.config.Config.User != "changed:user" {
+		t.Errorf("config.Config.USer was not updated! expected changed:user, got %s", mutator.config.Config.User)
+	}
+
+	// Check history.
+	if len(mutator.config.History) == 2 {
+		t.Errorf("config.History was changed")
+	}
+}
+
 func walkDescriptorRoot(ctx context.Context, engine casext.Engine, root ispec.Descriptor) (casext.DescriptorPath, error) {
 	var foundPath *casext.DescriptorPath
 
@@ -515,7 +580,7 @@ func TestMutatePath(t *testing.T) {
 		config.Labels["org.opensuse.testidx"] = label
 
 		// Update it.
-		if err := mutator.Set(context.Background(), config, meta, nil, ispec.History{
+		if err := mutator.Set(context.Background(), config, meta, nil, &ispec.History{
 			Comment: "change label " + label,
 		}); err != nil {
 			t.Fatalf("%d: unexpected error modifying config: %+v", idx, err)
