@@ -19,6 +19,7 @@ package layer
 
 import (
 	"archive/tar"
+	"bytes"
 	"fmt"
 	"io"
 	"os"
@@ -40,7 +41,8 @@ import (
 // would hope) after we check it the first time.
 var inUserNamespace = shared.RunningInUserNS()
 
-type tarExtractor struct {
+// TarExtractor represents a tar file to be extracted.
+type TarExtractor struct {
 	// mapOptions is the set of mapping options to use when extracting
 	// filesystem layers.
 	mapOptions MapOptions
@@ -55,7 +57,7 @@ type tarExtractor struct {
 	fsEval fseval.FsEval
 
 	// upperPaths are paths that have either been extracted in the execution of
-	// this tarExtractor or are ancestors of paths extracted. The purpose of
+	// this TarExtractor or are ancestors of paths extracted. The purpose of
 	// having this stored in-memory is to be able to handle opaque whiteouts as
 	// well as some other possible ordering issues with malformed archives (the
 	// downside of this approach is that it takes up memory -- we could switch
@@ -64,14 +66,14 @@ type tarExtractor struct {
 	upperPaths map[string]struct{}
 }
 
-// newTarExtractor creates a new tarExtractor.
-func newTarExtractor(opt MapOptions) *tarExtractor {
+// NewTarExtractor creates a new TarExtractor.
+func NewTarExtractor(opt MapOptions) *TarExtractor {
 	fsEval := fseval.DefaultFsEval
 	if opt.Rootless {
 		fsEval = fseval.RootlessFsEval
 	}
 
-	return &tarExtractor{
+	return &TarExtractor{
 		mapOptions:      opt,
 		partialRootless: opt.Rootless || inUserNamespace,
 		fsEval:          fsEval,
@@ -82,7 +84,7 @@ func newTarExtractor(opt MapOptions) *tarExtractor {
 // restoreMetadata applies the state described in tar.Header to the filesystem
 // at the given path. No sanity checking is done of the tar.Header's pathname
 // or other information. In addition, no mapping is done of the header.
-func (te *tarExtractor) restoreMetadata(path string, hdr *tar.Header) error {
+func (te *TarExtractor) restoreMetadata(path string, hdr *tar.Header) error {
 	// Some of the tar.Header fields don't match the OS API.
 	fi := hdr.FileInfo()
 
@@ -129,18 +131,54 @@ func (te *tarExtractor) restoreMetadata(path string, hdr *tar.Header) error {
 	// Apply xattrs. In order to make sure that we *only* have the xattr set we
 	// want, we first clear the set of xattrs from the file then apply the ones
 	// set in the tar.Header.
-	if err := te.fsEval.Lclearxattrs(path); err != nil {
+	if err := te.fsEval.Lclearxattrs(path, ignoreXattrs); err != nil {
 		return errors.Wrapf(err, "clear xattr metadata: %s", path)
 	}
 	for name, value := range hdr.Xattrs {
-		if err := te.fsEval.Lsetxattr(path, name, []byte(value), 0); err != nil {
+		value := []byte(value)
+
+		// Forbidden xattrs should never be touched.
+		if _, skip := ignoreXattrs[name]; skip {
+			// If the xattr is already set to the requested value, don't bail.
+			// The reason for this logic is kinda convoluted, but effectively
+			// because restoreMetadata is called with the *on-disk* metadata we
+			// run the risk of things like "security.selinux" being included in
+			// that metadata (and thus tripping the forbidden xattr error). By
+			// only touching xattrs that have a different value we are somewhat
+			// more efficient and we don't have to special case parent restore.
+			// Of course this will only ever impact ignoreXattrs.
+			if oldValue, err := te.fsEval.Lgetxattr(path, name); err == nil {
+				if bytes.Equal(value, oldValue) {
+					log.Debugf("restore xattr metadata: skipping already-set xattr %q: %s", name, hdr.Name)
+					continue
+				}
+			}
+			if te.partialRootless {
+				log.Warnf("rootless{%s} ignoring forbidden xattr: %q", hdr.Name, name)
+				continue
+			}
+			return errors.Errorf("restore xattr metadata: saw forbidden xattr %q: %s", name, hdr.Name)
+		}
+		if err := te.fsEval.Lsetxattr(path, name, value, 0); err != nil {
 			// In rootless mode, some xattrs will fail (security.capability).
 			// This is _fine_ as long as we're not running as root (in which
 			// case we shouldn't be ignoring xattrs that we were told to set).
+			//
+			// TODO: We should translate all security.capability capabilites
+			//       into v3 capabilities, which allow us to write them as
+			//       unprivileged users (we also would need to translate them
+			//       back when creating archives).
 			if te.partialRootless && os.IsPermission(errors.Cause(err)) {
 				log.Warnf("rootless{%s} ignoring (usually) harmless EPERM on setxattr %q", hdr.Name, name)
 				continue
 			}
+			// We cannot do much if we get an ENOTSUP -- this usually means
+			// that extended attributes are simply unsupported by the
+			// underlying filesystem (such as AUFS or NFS).
+			if errors.Cause(err) == unix.ENOTSUP {
+				log.Warnf("xatt{%s} ignoring ENOTSUP on setxattr %q", hdr.Name, name)
+				continue
+			}
 			return errors.Wrapf(err, "restore xattr metadata: %s", path)
 		}
 	}
@@ -153,11 +191,11 @@ func (te *tarExtractor) restoreMetadata(path string, hdr *tar.Header) error {
 }
 
 // applyMetadata applies the state described in tar.Header to the filesystem at
-// the given path, using the state of the tarExtractor to remap information
+// the given path, using the state of the TarExtractor to remap information
 // within the header. This should only be used with headers from a tar layer
 // (not from the filesystem). No sanity checking is done of the tar.Header's
 // pathname or other information.
-func (te *tarExtractor) applyMetadata(path string, hdr *tar.Header) error {
+func (te *TarExtractor) applyMetadata(path string, hdr *tar.Header) error {
 	// Modify the header.
 	if err := unmapHeader(hdr, te.mapOptions); err != nil {
 		return errors.Wrap(err, "unmap header")
@@ -167,11 +205,55 @@ func (te *tarExtractor) applyMetadata(path string, hdr *tar.Header) error {
 	return te.restoreMetadata(path, hdr)
 }
 
-// unpackEntry extracts the given tar.Header to the provided root, ensuring
+// isDirlink returns whether the given path is a link to a directory (or a
+// dirlink in rsync(1) parlance) which is used by --keep-dirlink to see whether
+// we should extract through the link or clobber the link with a directory (in
+// the case where we see a directory to extract and a symlink already exists
+// there).
+func (te *TarExtractor) isDirlink(root string, path string) (bool, error) {
+	// Make sure it exists and is a symlink.
+	if _, err := te.fsEval.Readlink(path); err != nil {
+		return false, errors.Wrap(err, "read dirlink")
+	}
+
+	// Technically a string.TrimPrefix would also work...
+	unsafePath, err := filepath.Rel(root, path)
+	if err != nil {
+		return false, errors.Wrap(err, "get relative-to-root path")
+	}
+
+	// It should be noted that SecureJoin will evaluate all symlinks in the
+	// path, so we don't need to loop over it or anything like that. It'll just
+	// be done for us (in UnpackEntry only the dirname(3) is evaluated but here
+	// we evaluate the whole thing).
+	targetPath, err := securejoin.SecureJoinVFS(root, unsafePath, te.fsEval)
+	if err != nil {
+		// We hit a symlink loop -- which is fine but that means that this
+		// cannot be considered a dirlink.
+		if errno := InnerErrno(err); errno == unix.ELOOP {
+			err = nil
+		}
+		return false, errors.Wrap(err, "sanitize old target")
+	}
+
+	targetInfo, err := te.fsEval.Lstat(targetPath)
+	if err != nil {
+		// ENOENT or similar just means that it's a broken symlink, which
+		// means we have to overwrite it (but it's an allowed case).
+		if securejoin.IsNotExist(err) {
+			err = nil
+		}
+		return false, err
+	}
+
+	return targetInfo.IsDir(), nil
+}
+
+// UnpackEntry extracts the given tar.Header to the provided root, ensuring
 // that the layer state is consistent with the layer state that produced the
 // tar archive being iterated over. This does handle whiteouts, so a tar.Header
 // that represents a whiteout will result in the path being removed.
-func (te *tarExtractor) unpackEntry(root string, hdr *tar.Header, r io.Reader) (Err error) {
+func (te *TarExtractor) UnpackEntry(root string, hdr *tar.Header, r io.Reader) (Err error) {
 	// Make the paths safe.
 	hdr.Name = CleanPath(hdr.Name)
 	root = filepath.Clean(root)
@@ -205,6 +287,7 @@ func (te *tarExtractor) unpackEntry(root string, hdr *tar.Header, r io.Reader) (
 	// directory will be fixed by later archive entries.
 	if dirFi, err := te.fsEval.Lstat(dir); err == nil && path != dir {
 		// FIXME: This is really stupid.
+		// #nosec G104
 		link, _ := te.fsEval.Readlink(dir)
 		dirHdr, err := tar.FileInfoHeader(dirFi, link)
 		if err != nil {
@@ -258,75 +341,83 @@ func (te *tarExtractor) unpackEntry(root string, hdr *tar.Header, r io.Reader) (
 		isOpaque := file == whOpaque
 		file = strings.TrimPrefix(file, whPrefix)
 
-		// We have special handling for opaque whiteouts. All other brands of
-		// whiteouts are just removed without prejudice (with the note that we
-		// cannot error out if a layer removes a non-existant file with this
-		// implementation -- in future we could add lowerPaths that would help
-		// track whether another whiteout caused the removal to "fail" or if
-		// the path was actually missing -- which would allow us to actually
-		// error out here).
+		// We have to be quite careful here. While the most intuitive way of
+		// handling whiteouts would be to just RemoveAll without prejudice, We
+		// have to be careful here. If there is a whiteout entry for a file
+		// *after* a normal entry (in the same layer) then the whiteout must
+		// not remove the new entry. We handle this by keeping track of
+		// whichpaths have been touched by this layer's extraction (these form
+		// the "upperdir"). We also have to handle cases where a directory has
+		// been marked for deletion, but a child has been extracted in this
+		// layer.
 
 		path = filepath.Join(dir, file)
 		if isOpaque {
 			path = dir
 		}
 
-		removeFunc := te.fsEval.RemoveAll
-		if isOpaque {
-			removeFunc = func(path string) error {
-				// Check that root exists.
-				if fi, err := te.fsEval.Lstat(path); err != nil {
-					return errors.Wrap(err, "check whiteout root")
-				} else if !fi.IsDir() {
-					return errors.Errorf("expected whiteout root to be directory: %v", path)
+		// If the root doesn't exist we've got nothing to do.
+		// XXX: We currently cannot error out if a layer asks us to remove a
+		//      non-existent path with this implementation (because we don't
+		//      know if it was implicitly removed by another whiteout). In
+		//      future we could add lowerPaths that would help track whether
+		//      another whiteout caused the removal to "fail" or if the path
+		//      was actually missing -- which would allow us to actually error
+		//      out here if the layer is invalid).
+		if _, err := te.fsEval.Lstat(path); err != nil {
+			// Need to use securejoin.IsNotExist to handle ENOTDIR.
+			if securejoin.IsNotExist(err) {
+				err = nil
+			}
+			return errors.Wrap(err, "check whiteout target")
+		}
+
+		// Walk over the path to remove it. We remove a given path as soon as
+		// it isn't present in upperPaths (which includes ancestors of paths
+		// we've extracted so we only need to look up the one path). Otherwise
+		// we iterate over any children and try again. The only difference
+		// between opaque whiteouts and regular whiteouts is that we don't
+		// delete the directory itself with opaque whiteouts.
+		err = te.fsEval.Walk(path, func(subpath string, info os.FileInfo, err error) error {
+			// If we are passed an error, bail unless it's ENOENT.
+			if err != nil {
+				// If something was deleted outside of our knowledge it's not
+				// the end of the world. In principle this shouldn't happen
+				// though, so we log it for posterity.
+				if os.IsNotExist(errors.Cause(err)) {
+					log.Debugf("whiteout removal hit already-deleted path: %s", subpath)
+					err = filepath.SkipDir
+				}
+				return err
+			}
+
+			// Get the relative form of subpath to root to match
+			// te.upperPaths.
+			upperPath, err := filepath.Rel(root, subpath)
+			if err != nil {
+				return errors.Wrap(err, "find relative-to-root [should never happen]")
+			}
+
+			// Remove the path only if it hasn't been touched.
+			if _, ok := te.upperPaths[upperPath]; !ok {
+				// Opaque whiteouts don't remove the directory itself, so skip
+				// the top-level directory.
+				if isOpaque && CleanPath(path) == CleanPath(subpath) {
+					return nil
 				}
 
-				// Walk over the path to remove children.
-				err := te.fsEval.Walk(path, func(subpath string, info os.FileInfo, err error) error {
-					// If we are passed an error, bail unless it's ENOENT.
-					if err != nil {
-						// If something was deleted outside of our knowledge
-						// it's not the end of the world. We've already checked
-						// that the root path exists.
-						if os.IsNotExist(errors.Cause(err)) {
-							err = filepath.SkipDir
-						}
-						return err
-					}
-
-					// Skip the top-level dir.
-					if CleanPath(path) == CleanPath(subpath) {
-						return nil
-					}
-
-					// Get the relative form of subpath to root to match
-					// te.upperPaths.
-					upperPath, err := filepath.Rel(root, subpath)
-					if err != nil {
-						return errors.Wrap(err, "find relative-to-root [should never happen]")
-					}
-
-					// Remove the path only if it hasn't been touched.
-					if _, ok := te.upperPaths[upperPath]; !ok {
-						err := errors.Wrap(te.fsEval.RemoveAll(subpath), "whiteout subpath")
-						// Skip anything underneath the subpath if it's a
-						// directory, since we just purged it.
-						if err == nil && info.IsDir() {
-							err = filepath.SkipDir
-						}
-						return err
-					}
-					return nil
-				})
-				return errors.Wrap(err, "opaque whiteout")
+				// Purge the path. We skip anything underneath (if it's a
+				// directory) since we just purged it -- and we don't want to
+				// hit ENOENT during iteration for no good reason.
+				err := errors.Wrap(te.fsEval.RemoveAll(subpath), "whiteout subpath")
+				if err == nil && info.IsDir() {
+					err = filepath.SkipDir
+				}
+				return err
 			}
-		}
-
-		// Run the removal function now.
-		if err := removeFunc(path); err != nil {
-			return errors.Wrap(err, "whiteout remove")
-		}
-		return nil
+			return nil
+		})
+		return errors.Wrap(err, "whiteout remove")
 	}
 
 	// Get information about the path. This has to be done after we've dealt
@@ -349,6 +440,7 @@ func (te *tarExtractor) unpackEntry(root string, hdr *tar.Header, r io.Reader) (
 		return errors.Wrap(err, "mkdir parent")
 	}
 
+	isDirlink := false
 	// We remove whatever existed at the old path to clobber it so that
 	// creating a new path will not break. The only exception is if the path is
 	// a directory in both the layer and the current filesystem, in which case
@@ -362,8 +454,33 @@ func (te *tarExtractor) unpackEntry(root string, hdr *tar.Header, r io.Reader) (
 	// "lower" file still exists (so the hard-link would point to the old
 	// inode). It's not clear if such an archive is actually valid though.
 	if !fi.IsDir() || hdr.Typeflag != tar.TypeDir {
-		if err := te.fsEval.RemoveAll(path); err != nil {
-			return errors.Wrap(err, "clobber old path")
+		// If we are in --keep-dirlinks mode and the existing fs object is a
+		// symlink to a directory (with the pending object is a directory), we
+		// don't remove the symlink (and instead allow subsequent objects to be
+		// just written through the symlink into the directory). This is a very
+		// specific usecase where layers that were generated independently from
+		// each other (on different base filesystems) end up with weird things
+		// like /lib64 being a symlink only sometimes but you never want to
+		// delete libraries (not just the ones that were under the "real"
+		// directory).
+		//
+		// TODO: This code should also handle a pending symlink entry where the
+		//       existing object is a directory. I'm not sure how we could
+		//       disambiguate this from a symlink-to-a-file but I imagine that
+		//       this is something that would also be useful in the same vein
+		//       as --keep-dirlinks (which currently only prevents clobbering
+		//       in the opposite case).
+		if te.mapOptions.KeepDirlinks &&
+			fi.Mode()&os.ModeSymlink == os.ModeSymlink && hdr.Typeflag == tar.TypeDir {
+			isDirlink, err = te.isDirlink(root, path)
+			if err != nil {
+				return errors.Wrap(err, "check is dirlink")
+			}
+		}
+		if !(isDirlink && te.mapOptions.KeepDirlinks) {
+			if err := te.fsEval.RemoveAll(path); err != nil {
+				return errors.Wrap(err, "clobber old path")
+			}
 		}
 	}
 
@@ -382,17 +499,25 @@ func (te *tarExtractor) unpackEntry(root string, hdr *tar.Header, r io.Reader) (
 		defer fh.Close()
 
 		// We need to make sure that we copy all of the bytes.
-		if n, err := io.Copy(fh, r); err != nil {
-			return err
-		} else if int64(n) != hdr.Size {
-			return errors.Wrap(io.ErrShortWrite, "unpack to regular file")
+		n, err := io.Copy(fh, r)
+		if int64(n) != hdr.Size {
+			err = io.ErrShortWrite
+		}
+		if err != nil {
+			return errors.Wrap(err, "unpack to regular file")
 		}
 
 		// Force close here so that we don't affect the metadata.
-		fh.Close()
+		if err := fh.Close(); err != nil {
+			return errors.Wrap(err, "close unpacked regular file")
+		}
 
 	// directory
 	case tar.TypeDir:
+		if isDirlink {
+			break
+		}
+
 		// Attempt to create the directory. We do a MkdirAll here because even
 		// though you need to have a tar entry for every component of a new
 		// path, applyMetadata will correct any inconsistencies.
