@@ -25,6 +25,7 @@ import (
 	"io/ioutil"
 	"os"
 	"path/filepath"
+	"strings"
 	"testing"
 	"time"
 
@@ -67,9 +68,9 @@ func testUnpackEntrySanitiseHelper(t *testing.T, dir, file, prefix string) func(
 			ChangeTime: time.Now(),
 		}
 
-		te := newTarExtractor(MapOptions{})
-		if err := te.unpackEntry(rootfs, hdr, bytes.NewBuffer(ctrValue)); err != nil {
-			t.Fatalf("unexpected unpackEntry error: %s", err)
+		te := NewTarExtractor(MapOptions{})
+		if err := te.UnpackEntry(rootfs, hdr, bytes.NewBuffer(ctrValue)); err != nil {
+			t.Fatalf("unexpected UnpackEntry error: %s", err)
 		}
 
 		hostValueGot, err := ioutil.ReadFile(filepath.Join(dir, "file"))
@@ -159,7 +160,7 @@ func TestUnpackEntrySymlinkScoping(t *testing.T) {
 	}(t)
 }
 
-// TestUnpackEntryParentDir ensures that when unpackEntry hits a path that
+// TestUnpackEntryParentDir ensures that when UnpackEntry hits a path that
 // doesn't have its leading directories, we create all of the parent
 // directories.
 func TestUnpackEntryParentDir(t *testing.T) {
@@ -190,9 +191,9 @@ func TestUnpackEntryParentDir(t *testing.T) {
 		ChangeTime: time.Now(),
 	}
 
-	te := newTarExtractor(MapOptions{})
-	if err := te.unpackEntry(rootfs, hdr, bytes.NewBuffer(ctrValue)); err != nil {
-		t.Fatalf("unexpected unpackEntry error: %s", err)
+	te := NewTarExtractor(MapOptions{})
+	if err := te.UnpackEntry(rootfs, hdr, bytes.NewBuffer(ctrValue)); err != nil {
+		t.Fatalf("unexpected UnpackEntry error: %s", err)
 	}
 
 	ctrValueGot, err := ioutil.ReadFile(filepath.Join(rootfs, "a/b/c/file"))
@@ -277,9 +278,9 @@ func TestUnpackEntryWhiteout(t *testing.T) {
 				Typeflag: tar.TypeReg,
 			}
 
-			te := newTarExtractor(MapOptions{})
-			if err := te.unpackEntry(dir, hdr, nil); err != nil {
-				t.Fatalf("unexpected error in unpackEntry: %s", err)
+			te := NewTarExtractor(MapOptions{})
+			if err := te.UnpackEntry(dir, hdr, nil); err != nil {
+				t.Fatalf("unexpected error in UnpackEntry: %s", err)
 			}
 
 			// Make sure that the path is gone.
@@ -441,6 +442,31 @@ func TestUnpackOpaqueWhiteout(t *testing.T) {
 				{"another/dir", "", tar.TypeDir, false},
 				{"another/dir/somewhere", "", tar.TypeReg, false},
 			}},
+			{"UpperWhiteout", false, []pseudoHdr{
+				{whPrefix + "fileB", "", tar.TypeReg, true},
+				{"fileA", "", tar.TypeReg, true},
+				{"fileB", "", tar.TypeReg, true},
+				{"fileC", "", tar.TypeReg, false},
+				{whPrefix + "fileA", "", tar.TypeReg, true},
+				{whPrefix + "fileC", "", tar.TypeReg, true},
+			}},
+			// XXX: What umoci should do here is not really defined by the
+			//      spec. In particular, whether you need a whiteout for every
+			//      sub-path or just the path itself is not well-defined. This
+			//      code assumes that you *do not*.
+			{"UpperDirWhiteout", false, []pseudoHdr{
+				{whPrefix + "dir2", "", tar.TypeReg, true},
+				{"file", "", tar.TypeReg, false},
+				{"dir1", "", tar.TypeDir, true},
+				{"dir1/file", "", tar.TypeRegA, true},
+				{"dir1/link", "../badlink", tar.TypeSymlink, false},
+				{"dir1/verybadlink", "../../../../../../../../../../../../etc/shadow", tar.TypeSymlink, true},
+				{"dir1/verybadlink2", "/../../../../../../../../../../../../etc/shadow", tar.TypeSymlink, false},
+				{whPrefix + "dir1", "", tar.TypeReg, true},
+				{"dir2", "", tar.TypeDir, true},
+				{"dir2/file", "", tar.TypeRegA, true},
+				{"dir2/link", "../badlink", tar.TypeSymlink, false},
+			}},
 		} {
 			t.Logf("running Test%s", test.name)
 			mapOptions := MapOptions{
@@ -463,8 +489,8 @@ func TestUnpackOpaqueWhiteout(t *testing.T) {
 			// Track if we have upper entries.
 			numUpper := 0
 
-			// First we apply the non-upper files in a new tarExtractor.
-			te := newTarExtractor(mapOptions)
+			// First we apply the non-upper files in a new TarExtractor.
+			te := NewTarExtractor(mapOptions)
 			for _, ph := range test.pseudoHeaders {
 				// Skip upper.
 				if ph.upper {
@@ -473,13 +499,13 @@ func TestUnpackOpaqueWhiteout(t *testing.T) {
 				}
 				hdr, rdr := fromPseudoHdr(ph)
 				hdr.Name = filepath.Join(whiteoutDir, hdr.Name)
-				if err := te.unpackEntry(dir, hdr, rdr); err != nil {
-					t.Errorf("unpackEntry %s failed: %v", hdr.Name, err)
+				if err := te.UnpackEntry(dir, hdr, rdr); err != nil {
+					t.Errorf("UnpackEntry %s failed: %v", hdr.Name, err)
 				}
 			}
 
-			// Now we apply the upper files in another tarExtractor.
-			te = newTarExtractor(mapOptions)
+			// Now we apply the upper files in another TarExtractor.
+			te = NewTarExtractor(mapOptions)
 			for _, ph := range test.pseudoHeaders {
 				// Skip non-upper.
 				if !ph.upper {
@@ -487,8 +513,8 @@ func TestUnpackOpaqueWhiteout(t *testing.T) {
 				}
 				hdr, rdr := fromPseudoHdr(ph)
 				hdr.Name = filepath.Join(whiteoutDir, hdr.Name)
-				if err := te.unpackEntry(dir, hdr, rdr); err != nil {
-					t.Errorf("unpackEntry %s failed: %v", hdr.Name, err)
+				if err := te.UnpackEntry(dir, hdr, rdr); err != nil {
+					t.Errorf("UnpackEntry %s failed: %v", hdr.Name, err)
 				}
 			}
 
@@ -497,7 +523,7 @@ func TestUnpackOpaqueWhiteout(t *testing.T) {
 				Name:     filepath.Join(whiteoutDir, whOpaque),
 				Typeflag: tar.TypeReg,
 			}
-			if err := te.unpackEntry(dir, whHdr, nil); err != nil {
+			if err := te.UnpackEntry(dir, whHdr, nil); err != nil {
 				t.Errorf("unpack whiteout %s failed: %v", whiteoutRoot, err)
 				continue
 			}
@@ -506,8 +532,14 @@ func TestUnpackOpaqueWhiteout(t *testing.T) {
 			// should have survived. If it was in lower it shouldn't. We don't
 			// bother checking the contents here.
 			for _, ph := range test.pseudoHeaders {
-				fullPath := filepath.Join(whiteoutRoot, ph.path)
+				// If there's an explicit whiteout in the headers we ignore it
+				// here, since it won't be on the filesystem.
+				if strings.HasPrefix(filepath.Base(ph.path), whPrefix) {
+					t.Logf("ignoring whiteout entry %q during post-check", ph.path)
+					continue
+				}
 
+				fullPath := filepath.Join(whiteoutRoot, ph.path)
 				_, err := te.fsEval.Lstat(fullPath)
 				if err != nil && !os.IsNotExist(errors.Cause(err)) {
 					t.Errorf("unexpected lstat error of %s: %v", ph.path, err)
@@ -565,7 +597,7 @@ func TestUnpackHardlink(t *testing.T) {
 		hardFileB = "hard link to symlink"
 	)
 
-	te := newTarExtractor(MapOptions{})
+	te := NewTarExtractor(MapOptions{})
 
 	// Regular file.
 	hdr = &tar.Header{
@@ -579,8 +611,8 @@ func TestUnpackHardlink(t *testing.T) {
 		AccessTime: time.Now(),
 		ChangeTime: time.Now(),
 	}
-	if err := te.unpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
-		t.Fatalf("regular: unexpected unpackEntry error: %s", err)
+	if err := te.UnpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
+		t.Fatalf("regular: unexpected UnpackEntry error: %s", err)
 	}
 
 	// Hardlink to regFile.
@@ -592,8 +624,8 @@ func TestUnpackHardlink(t *testing.T) {
 		Uid: os.Getuid() + 1337,
 		Gid: os.Getgid() + 2020,
 	}
-	if err := te.unpackEntry(dir, hdr, nil); err != nil {
-		t.Fatalf("hardlinkA: unexpected unpackEntry error: %s", err)
+	if err := te.UnpackEntry(dir, hdr, nil); err != nil {
+		t.Fatalf("hardlinkA: unexpected UnpackEntry error: %s", err)
 	}
 
 	// Symlink to regFile.
@@ -604,8 +636,8 @@ func TestUnpackHardlink(t *testing.T) {
 		Typeflag: tar.TypeSymlink,
 		Linkname: filepath.Join("../../../", regFile),
 	}
-	if err := te.unpackEntry(dir, hdr, nil); err != nil {
-		t.Fatalf("symlink: unexpected unpackEntry error: %s", err)
+	if err := te.UnpackEntry(dir, hdr, nil); err != nil {
+		t.Fatalf("symlink: unexpected UnpackEntry error: %s", err)
 	}
 
 	// Hardlink to symlink.
@@ -617,8 +649,8 @@ func TestUnpackHardlink(t *testing.T) {
 		Uid: os.Getuid() + 1337,
 		Gid: os.Getgid() + 2020,
 	}
-	if err := te.unpackEntry(dir, hdr, nil); err != nil {
-		t.Fatalf("hardlinkB: unexpected unpackEntry error: %s", err)
+	if err := te.UnpackEntry(dir, hdr, nil); err != nil {
+		t.Fatalf("hardlinkB: unexpected UnpackEntry error: %s", err)
 	}
 
 	// Quickly make sure that the contents are as expected.
@@ -725,7 +757,7 @@ func TestUnpackEntryMap(t *testing.T) {
 				symDir   = "link-dir"
 			)
 
-			te := newTarExtractor(MapOptions{
+			te := NewTarExtractor(MapOptions{
 				UIDMappings: []rspec.LinuxIDMapping{test.uidMap},
 				GIDMappings: []rspec.LinuxIDMapping{test.gidMap},
 			})
@@ -743,8 +775,8 @@ func TestUnpackEntryMap(t *testing.T) {
 				AccessTime: time.Now(),
 				ChangeTime: time.Now(),
 			}
-			if err := te.unpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
-				t.Fatalf("regfile: unexpected unpackEntry error: %s", err)
+			if err := te.UnpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
+				t.Fatalf("regfile: unexpected UnpackEntry error: %s", err)
 			}
 
 			if err := unix.Lstat(filepath.Join(dir, hdr.Name), &fi); err != nil {
@@ -772,8 +804,8 @@ func TestUnpackEntryMap(t *testing.T) {
 				AccessTime: time.Now(),
 				ChangeTime: time.Now(),
 			}
-			if err := te.unpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
-				t.Fatalf("regdir: unexpected unpackEntry error: %s", err)
+			if err := te.UnpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
+				t.Fatalf("regdir: unexpected UnpackEntry error: %s", err)
 			}
 
 			if err := unix.Lstat(filepath.Join(dir, hdr.Name), &fi); err != nil {
@@ -801,8 +833,8 @@ func TestUnpackEntryMap(t *testing.T) {
 				AccessTime: time.Now(),
 				ChangeTime: time.Now(),
 			}
-			if err := te.unpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
-				t.Fatalf("regdir: unexpected unpackEntry error: %s", err)
+			if err := te.UnpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
+				t.Fatalf("regdir: unexpected UnpackEntry error: %s", err)
 			}
 
 			if err := unix.Lstat(filepath.Join(dir, hdr.Name), &fi); err != nil {
@@ -830,8 +862,8 @@ func TestUnpackEntryMap(t *testing.T) {
 				AccessTime: time.Now(),
 				ChangeTime: time.Now(),
 			}
-			if err := te.unpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
-				t.Fatalf("regdir: unexpected unpackEntry error: %s", err)
+			if err := te.UnpackEntry(dir, hdr, bytes.NewBuffer(ctrValue)); err != nil {
+				t.Fatalf("regdir: unexpected UnpackEntry error: %s", err)
 			}
 
 			if err := unix.Lstat(filepath.Join(dir, hdr.Name), &fi); err != nil {
@@ -849,3 +881,67 @@ func TestUnpackEntryMap(t *testing.T) {
 		}
 	}(t)
 }
+
+func TestIsDirlink(t *testing.T) {
+	dir, err := ioutil.TempDir("", "umoci-TestDirLink")
+	if err != nil {
+		t.Fatal(err)
+	}
+	defer os.RemoveAll(dir)
+
+	if err := os.Mkdir(filepath.Join(dir, "test_dir"), 0755); err != nil {
+		t.Fatal(err)
+	}
+	if file, err := os.Create(filepath.Join(dir, "test_file")); err != nil {
+		t.Fatal(err)
+	} else {
+		file.Close()
+	}
+	if err := os.Symlink("test_dir", filepath.Join(dir, "link")); err != nil {
+		t.Fatal(err)
+	}
+
+	te := NewTarExtractor(MapOptions{})
+	// Basic symlink usage.
+	if dirlink, err := te.isDirlink(dir, filepath.Join(dir, "link")); err != nil {
+		t.Errorf("symlink failed: %v", err)
+	} else if !dirlink {
+		t.Errorf("dirlink test failed")
+	}
+
+	// "Read" a non-existent link.
+	if _, err := te.isDirlink(dir, filepath.Join(dir, "doesnt-exist")); err == nil {
+		t.Errorf("read non-existent dirlink")
+	}
+	// "Read" a directory.
+	if _, err := te.isDirlink(dir, filepath.Join(dir, "test_dir")); err == nil {
+		t.Errorf("read non-link dirlink")
+	}
+	// "Read" a file.
+	if _, err := te.isDirlink(dir, filepath.Join(dir, "test_file")); err == nil {
+		t.Errorf("read non-link dirlink")
+	}
+
+	// Break the symlink.
+	if err := os.Remove(filepath.Join(dir, "test_dir")); err != nil {
+		t.Fatal(err)
+	}
+	if dirlink, err := te.isDirlink(dir, filepath.Join(dir, "link")); err != nil {
+		t.Errorf("broken symlink failed: %v", err)
+	} else if dirlink {
+		t.Errorf("broken dirlink test failed")
+	}
+
+	// Point the symlink to a file.
+	if err := os.Remove(filepath.Join(dir, "link")); err != nil {
+		t.Fatal(err)
+	}
+	if err := os.Symlink("test_file", filepath.Join(dir, "link")); err != nil {
+		t.Fatal(err)
+	}
+	if dirlink, err := te.isDirlink(dir, filepath.Join(dir, "link")); err != nil {
+		t.Errorf("file symlink failed: %v", err)
+	} else if dirlink {
+		t.Errorf("file dirlink test failed")
+	}
+}
