@@ -34,6 +34,8 @@ import (
 // fiRestore restores the state given by an os.FileInfo instance at the given
 // path by ensuring that an Lstat(path) will return as-close-to the same
 // os.FileInfo.
+//
+// #nosec G104
 func fiRestore(path string, fi os.FileInfo) {
 	// archive/tar handles the OS-specific syscall stuff required to get atime
 	// and mtime information for a file.
@@ -330,7 +332,8 @@ func foreachSubpath(path string, wrapFn WrapFunc) error {
 	// permissions because we're already in a context with filepath.Dir(path)
 	// is at least a+rx. However, because we are calling wrapFn we need to
 	// restore the original mode immediately.
-	os.Chmod(path, fi.Mode()|0444)
+	// #nosec G104
+	_ = os.Chmod(path, fi.Mode()|0444)
 	names, err := fd.Readdirnames(-1)
 	fiRestore(path, fi)
 	if err != nil {
@@ -507,13 +510,16 @@ func Lgetxattr(path, name string) ([]byte, error) {
 // properly all of the internal functions were wrapped with unpriv.Wrap to make
 // it possible to create a path even if you do not currently have enough access
 // bits.
-func Lclearxattrs(path string) error {
+func Lclearxattrs(path string, except map[string]struct{}) error {
 	return errors.Wrap(Wrap(path, func(path string) error {
 		names, err := Llistxattr(path)
 		if err != nil {
 			return err
 		}
 		for _, name := range names {
+			if _, skip := except[name]; skip {
+				continue
+			}
 			if err := Lremovexattr(path, name); err != nil {
 				// SELinux won't let you change security.selinux (for obvious
 				// security reasons), so we don't clear xattrs if attempting to
